unit FTP400;

{
  FTP for AS/400 unit for Delphi 32
  Copyright (C) 2001  Paul Toth <tothpaul@free.fr>
  http://tothpaul.free.fr

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

}


// v1.0 ????-??-??
// v1.1 2001-10-25, TFile.Describe fix by Holgen


interface

uses
 Windows,Forms,Classes,StdCtrls,SysUtils,CrtSock;

type
  TFile=class; // forward

  TField=class
  private
   owner  :TFile;
   fName  :string;
   text   :string;
   kind   :char;
   size   :integer;
   digits :integer;
   decimal:integer;
   offset :integer;
  protected
   Constructor Create(AOwner:TFile;FD:string);
   Procedure Convert(Var s:string);
  public
   function Width:integer;
   function AsString:string;
   property Name:string read fName;
  end;

  TDownload=class
  private
   Handle:integer;
   sin,sout:TextFile;
   Pasv:integer;
   addr:string;
   Port:word;
   pin,pout:TextFile;
   CacheName:string;
   Cache:TFileStream;
   CacheSize:integer;
   CacheCount:integer;
   Text:TStringList;
   fMemo:TMemo;
   fHost:string;
   fUser:string;
   fPassword:string;
  protected
   procedure CacheAdd(s:string); virtual;
   procedure Log(s:string);
   function ReadString:string;
   procedure WriteString(s:string);
   function Status:char;
   function Exec(s:string):char;
   function Echo(s:string):string;
   function Login:boolean;
   function Passive:boolean;
   function GetFile(FileName:string;RecSize:integer):boolean;
   function ReceiveData(RecSize:integer):boolean;
  public
   constructor Create; virtual;
   destructor Destroy; override;
  published
   property Memo:TMemo read fMemo write fMemo;
   property Host:string read fHost write fHost;
   property User:string read fUser write fUser;
   property Password:string read fPassword write fPassword;
  end;

  TLibList=class(TDownload)
  public
   Function GetList(List:TStrings):boolean;
  end;

  TFileList=class(TDownload)
  public
   Function GetList(Lib:string; List:TStrings):boolean;
  end;

  TFile=class(TDownload)
  private
   Fields:TList;
   Cursor:integer;
   function GetField(index:integer):TField;
  protected
   procedure CacheAdd(s:string); override;
   procedure Clear;
   function Current:string;
   function Describe(FileName:string):boolean;
   function RecordSize:integer;
  public
   constructor Create; override;
   destructor Destroy; override;
   function Download(FileName:string):boolean;
   procedure First;
   procedure Last;
   procedure Next;
   procedure Prior;
   function Bof:boolean;
   function Eof:boolean;
   function RecordCount:integer;
   function FieldCount:integer;
   property Field[index:integer]:TField read GetField;
  end;

implementation

Constructor TDownload.Create;
 function TempFileName:string;
  var
   p:array[0..128] of char;
  begin
   GetTempFileName('.','FTP',0,p);
   Result:=p;
  end;
 begin
  Text:=TStringList.Create;
  CacheName:=TempFileName;
  Cache:=TFileStream.Create(CacheName,fmCreate or fmOpenReadWrite);
 end;

Destructor TDownload.Destroy;
 begin
  Text.Free;
  Cache.Position:=0;
  SetEndOfFile(Cache.Handle);
  Cache.Free;
  DeleteFile(CacheName);
  inherited Destroy;
 end;

procedure TDownLoad.CacheAdd(s:string);
 begin
  inc(CacheCount);
  Cache.Writebuffer(s[1],Length(s));
 end;

procedure TDownload.Log(s:string);
 begin
  if Assigned(fMemo) then fMemo.Lines.add(s);
 end;

function TDownload.ReadString:string;
 begin
  repeat
   ReadLn(sin,Result);
   Log(Result);
  until (length(Result)<4)or(Result[4]<>'-');
 end;

procedure TDownload.WriteString(s:string);
 begin
  Log(s);
  WriteLn(sout,s);
 end;

function TDownload.Status:char;
 var
  s:string;
 begin
  s:=ReadString; if s='' then s:='?';
  Result:=s[1];
 end;

function TDownload.Exec(s:string):char;
 begin
  WriteString(s);
  Result:=Status;
 end;

function TDownload.Echo(s:string):string;
 begin
  WriteString(s);
  Result:=ReadString;
 end;

function TDownload.Login:boolean;
 var
  s:string;
 begin
  Result:=(Status='2') and (Exec('USER '+fUser)='3') and (Exec('PASS '+fPassword)='2');
 end;

function TDownload.Passive:boolean;
 var
  p:integer;

  function GetValue:integer;
   var
    x:integer;
   begin
    x:=Length(Addr);
    while Addr[x]<>',' do dec(x);
    Result:=StrToInt(copy(Addr,x+1,length(Addr)));
    SetLength(Addr,x-1);
   end;

 begin
  Result:=False;
  Addr:=Echo('PASV'); if Addr[1]<>'2' then exit;
  Addr:=Trim(Copy(Addr,5,Length(Addr))); // ip,port
  p:=pos('(',Addr); if p>0 then Addr:=Copy(Addr,p+1,length(Addr)); SetLength(Addr,pos(')',Addr)-1);
  p:=GetValue; Port:=256*GetValue+p; // get Port
  for p:=1 to Length(Addr) do if Addr[p]=',' then Addr[p]:='.'; // replace "," by "." in IP address
  Result:=True;
 end;

function TDownload.GetFile(FileName:string;RecSize:integer):boolean;
 begin
  Result:=False;
  if (RecSize>0)and((Exec('SITE TRIM 0')<>'2')or(Exec('TYPE E')<>'2')) then exit;
  if not Passive then exit;
  WriteString('RETR '+FileName);
  Result:=ReceiveData(RecSize);
 end;

Function TDownload.ReceiveData(RecSize:integer):boolean;
 Var
  HData:integer;
  din,dout:TextFile;
  s,ss:string;
  p,i:integer;
 begin
  Cache.Position:=0;
  CacheSize:=RecSize;
  CacheCount:=0;
  Result:=False;
  HData:=CallServer(Host,Port);
  if hData=-1 then exit;
  if Status<>'1' then exit;
  if RecSize<0 then begin
   AssignCrtSock(hData,din,dout);
   ReadLn(din,s);
   Repeat
    Text.Add(s);
    ReadLn(din,s);
   until s='';
  end else begin
   s:='';
   SetLength(ss,RecSize);
   p:=recv(hData,pchar(ss),RecSize,0);
   while p>0 do begin
    i:=RecSize-Length(s);
    if i>p then begin // need more then available
     s:=s+copy(ss,1,p); // add to allready loaded data
    end else begin
     s:=s+Copy(ss,1,i); // add what we need
     CacheAdd(s);
     s:=Copy(ss,i+1,p-i); // get what left
    end;
    p:=recv(hData,pchar(ss),RecSize,0);
   end;
  end;
  Result:=(Status='2');
  Disconnect(hData);
 end;

//-----------

Function TLibList.GetList(List:TStrings):boolean;
 var
  i:integer;
 function LibList:boolean;
  begin
   Result:=False;
   // Echo('RCMD CHGJOB LOG(4 00 *SECLVL) LOGCLPGM(*YES)');
   if Exec('RCMD OVRPRTF QPRTLIBL HOLD(*YES) OVRSCOPE(*JOB)')<>'2' then exit;
   if Exec('RCMD DSPLIBL *PRINT')<>'2' then exit;
   if Exec('RCMD CRTPF QTEMP/LIBL RCDLEN(80)')<>'2' then exit;
   if Exec('RCMD CPYSPLF QPRTLIBL QTEMP/LIBL JOB('+fUser+'/QPRTJOB) SPLNBR(*LAST)')<>'2' then exit;
   if Exec('RCMD DLTOVR QPRTLIBL LVL(*JOB)')<>'2' then exit;
   if Exec('RCMD DLTSPLF QPRTLIBL JOB('+fUser+'/QPRTJOB) SPLNBR(*LAST)')<>'2' then exit;
   Result:=True;
  end;
 begin
  Text.Clear;
  List.Clear;
  Handle:=CallServer(fHost,21);
  Result:=(Handle<>-1);
  if Result then begin
   AssignCrtSock(Handle,sin,sout);
   Result:=Login and LibList and GetFile('QTEMP/LIBL',-1);
   if Result then
    for i:=2 to Text.Count-2 do List.Add(Trim(Copy(Text[i],4,10)));
   Exec('QUIT');
   Disconnect(Handle);
  end;
  Text.Clear;
 end;

//-----------

Function TFileList.GetList(Lib:string;List:TStrings):boolean;
 var
  i:integer;
  s:string;
  p:integer;
 function FileList:boolean;
  begin
   Result:=False;
   if not Passive then exit;
   WriteString('LIST '+Lib);
   Result:=ReceiveData(-1);
  end;
 begin
  List.Clear;
  Handle:=CallServer(fHost,21);
  Result:=(Handle<>-1);
  if Result then begin
   AssignCrtSock(Handle,sin,sout);
   Result:=Login and FileList;
   if Result then
    for i:=0 to Text.Count-1 do begin
     s:=Trim(Copy(Text[i],51,3*10+2));
     p:=pos('/',s); if p>0 then Delete(s,1,p);
     List.Add(s);
    end;
   Exec('QUIT');
   Disconnect(Handle);
  end;
 end;

//-----------

Constructor TFile.Create;
 begin
  inherited Create;
  Fields:=TList.Create;
  Cursor:=-1;
 end;

Destructor TFile.Destroy;
 begin
  Clear;
  Fields.Free;
  inherited Destroy;
 end;

Procedure TFile.CacheAdd(s:string);
 var
  i:integer;
 begin
  for i:=0 to Fields.Count-1 do TField(Fields[i]).Convert(s);
  inherited CacheAdd(s);
 end;

Procedure TFile.Clear;
 var
  i:integer;
 begin
  for i:=0 to Fields.Count-1 do TField(Fields[i]).Free; Fields.Clear;
  Text.Clear;
  Cache.Position:=0;
  Cursor:=-1;
 end;

Function TFile.GetField(index:integer):TField;
 begin
  Result:=Fields[Index];
 end;

Function TFile.Current:string;
 begin
  Cache.Position:=CacheSize*Cursor;
  SetLength(Result,CacheSize);
  Cache.ReadBuffer(Result[1],CacheSize);
 end;

function TFile.Describe(FileName:string):boolean;
 var
  i:integer;
  p:integer;
  s:string;
 begin
  Result:=false;

  p:=pos('.',FileName);
  if p>0 then s:=Copy(FileName,1,p-1) else s:=FileName;

//-- fix for last OS400 Version (with binary data) by Holger <hwagenle@de.hellmann.net> 
// if Exec('RCMD DSPFFD '+s+' OUTPUT(*OUTFILE) OUTFILE(QTEMP/FFD)')<>'2' then exit;
  if Exec('RCMD DSPFFD '+s+' OUTPUT(*OUTFILE) OUTFILE(QTEMP/FFD1)')<>'2' then exit;
  if Exec('RCMD CRTPF QTEMP/FFD RCDLEN(385)')<>'2' then exit;
  if Exec('RCMD CPYF QTEMP/FFD1 QTEMP/FFD MBROPT(*REPLACE) FMTOPT(*NOCHK)') <>'2' then exit;
//-- end of fix
  if not GetFile('QTEMP/FFD',-1) then exit;
  if Text.Count=0 then exit;

  for i:=0 to Text.Count-1 do TField.Create(Self,Text[i]);
  // bad fix :( for SRCFILES
  if (Text.Count=3)and(Field[0].Name='SRCSEQ')and(Field[1].Name='SRCDAT')and(Field[2].Name='SRCDTA') then begin
   s:=Text[2];
   Clear;
   with TField.Create(Self,s) do Offset:=1;
  end;
  Text.Clear;
  Result:=True;
 end;

function TFile.RecordSize:integer;
 var
  i:integer;
 begin
  Result:=1;
  for i:=0 to Fields.Count-1 do inc(Result,TField(Fields[i]).Size);
 end;

function TFile.Download(FileName:string):boolean;
 var
  i,j:integer;
 begin
  Clear;
  Handle:=CallServer(fHost,21);
  Result:=(Handle<>-1);
  if Result then begin
   AssignCrtSock(Handle,sin,sout);
   Result:=Login and Describe(FileName) and GetFile(FileName,RecordSize);
   Exec('QUIT');
   Disconnect(Handle);
  end;
 end;

Procedure TFile.First;
 begin
  if CacheCount=0 then Cursor:=-1 else Cursor:=0;
 end;

Procedure TFile.Last;
 begin
  if CacheCount=0 then Cursor:=-1 else Cursor:=CacheCount-1;
 end;

Procedure TFile.Next;
 begin
  inc(Cursor);
  if Cursor>=CacheCount then Cursor:=-1;
 end;

Procedure TFile.Prior;
 begin
  if Cursor>=0 then Dec(Cursor);
 end;

Function TFile.Bof:boolean;
 begin
  Result:=(Cursor=0);
 end;

Function TFile.Eof:boolean;
 begin
  Result:=(Cursor=CacheCount-1);
 end;

Function TFile.RecordCount:integer;
 begin
  Result:=CacheCount;
 end;

Function TFile.FieldCount:integer;
 begin
  Result:=Fields.Count;
 end;

//--------------

Constructor TField.Create(AOwner:TFile;FD:string);
 begin
  Owner:=AOwner;
  Owner.Fields.Add(Self);
 // FD is a line from "DSPFFD OUTPUT(*OUTFILE)"
  fName  :=Trim(Copy(FD,130,10));
  Offset :=StrToInt(Copy(FD,150,5));
  Size   :=StrToInt(Copy(FD,160,5));
  Digits :=StrToInt(Copy(FD,165,2));
  Decimal:=StrToInt(Copy(FD,167,2));
  Text   :=Trim(Copy(FD,169,50));
  Kind   :=FD[322];
 end;

function StrUnpack(p:pchar;size:integer):string;
 var
  i:integer;
  b:byte;
  v:integer;
 begin
  Result:='';
  v:=0;
  for i:=0 to Size-2 do begin
   b:=ord(p[i]);
   v:=100*v+b and $F+10*(b shr 4);
  end;
  v:=10*v+ord(p[Size-1]) shr 4;
  Result:=Result+' '+IntToStr(v);
 end;

Const
 E2A:array[#0..#255] of char =(
 #$00,#$01,#$02,#$03,#$5F,#$09,#$2B,#$83,#$2D,#$5F,#$2B,#$0B,#$0C,#$0D,#$0E,#$A4,#$10,#$11,#$12,#$13, { 20}
 #$5F,#$2B,#$08,#$2D,#$18,#$19,#$2B,#$5F,#$1A,#$1D,#$1E,#$1F,#$2D,#$A6,#$2B,#$2B,#$2B,#$0A,#$17,#$1B, { 40}
 #$A6,#$2D,#$2B,#$5F,#$5F,#$05,#$06,#$07,#$2D,#$A6,#$16,#$2B,#$2B,#$2B,#$A6,#$04,#$A6,#$2D,#$2B,#$5F, { 60}
 #$5F,#$5F,#$5F,#$1C,#$20,#$5F,#$E2,#$E4,#$40,#$E1,#$E3,#$E5,#$5C,#$F1,#$B0,#$2E,#$3C,#$28,#$2B,#$21, { 80}
 #$26,#$7B,#$EA,#$EB,#$7D,#$ED,#$EE,#$EF,#$EC,#$DF,#$A7,#$24,#$2A,#$29,#$3B,#$5E,#$2D,#$2F,#$C2,#$C4, {100}
 #$C0,#$C1,#$C3,#$C5,#$C7,#$D1,#$F9,#$2C,#$25,#$5F,#$3E,#$3F,#$F8,#$C9,#$CA,#$CB,#$C8,#$CD,#$CE,#$CF, {120}
 #$CC,#$B5,#$3A,#$A3,#$E0,#$27,#$3D,#$22,#$D8,#$61,#$62,#$63,#$64,#$65,#$66,#$67,#$68,#$69,#$AB,#$BB, {140}
 #$F0,#$FD,#$FE,#$B1,#$5B,#$6A,#$6B,#$6C,#$6D,#$6E,#$6F,#$70,#$71,#$72,#$AA,#$BA,#$E6,#$2C,#$C6,#$A4, {160}
 #$27,#$22,#$73,#$74,#$75,#$76,#$77,#$78,#$79,#$7A,#$A1,#$BF,#$D0,#$DD,#$DE,#$AE,#$A2,#$23,#$A5,#$B7, {180}
 #$A9,#$5D,#$B6,#$BC,#$BD,#$BE,#$AC,#$7C,#$AF,#$7E,#$60,#$D7,#$E9,#$41,#$42,#$43,#$44,#$45,#$46,#$47, {200}
 #$48,#$49,#$2D,#$F4,#$F6,#$F2,#$F3,#$F5,#$E8,#$4A,#$4B,#$4C,#$4D,#$4E,#$4F,#$50,#$51,#$52,#$B9,#$FB, {220}
 #$FC,#$7C,#$FA,#$FF,#$E7,#$F7,#$53,#$54,#$55,#$56,#$57,#$58,#$59,#$5A,#$B2,#$D4,#$D6,#$D2,#$D3,#$D5, {240}
 #$30,#$31,#$32,#$33,#$34,#$35,#$36,#$37,#$38,#$39,#$B3,#$DB,#$DC,#$D9,#$DA,#$7F);                    {256}

Procedure TField.Convert(var s:string);
 var
  i:integer;
 begin
  case Kind of
   'P': exit;
  end;
  for i:=Offset to Offset+Size-1 do s[i]:=E2A[s[i]];
 end;

Function TField.Width:integer;
 begin
  if Kind='P' then
   Result:=2*Size
  else
   Result:=Size;
  if Result<Length(Name) then Result:=Length(Name);
 end;

Function TField.AsString:string;
 var
  s:string;
 begin
  s:=Owner.Current;
  if Kind='P' then
   Result:=StrUnpack(@s[Offset],Size)
  else
   Result:=Copy(s,Offset,Size);
 end;

end.
